"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.groupBy = exports.keyBy = exports.range = exports.isEqual = exports.map = exports.every = exports.some = exports.filter = exports.find = exports.findIndex = exports.merge = exports.cloneDeep = exports.clone = exports.flow = exports.uniqBy = exports.uniq = exports.once = exports.uniqueId = exports.compact = exports.flatten = exports.indexOf = exports.kebabCase = exports.forEach = exports.pickBy = exports.isEmpty = exports.has = exports.get = exports.at = exports.head = exports.take = exports.sortBy = void 0;
/**
 * Sort objects by object properties
 */
function sortBy(arr, keys) {
    return arr.slice().sort((a, b) => {
        for (let i = 0; i < keys.length; i++) {
            const key = keys[i];
            if (a[key] < b[key])
                return -1;
            else if (a[key] > b[key])
                return 1;
        }
        return 0;
    });
}
exports.sortBy = sortBy;
/**
 * Take out the first N items of an array
 */
function take(arr, n) {
    return arr.slice(0, n);
}
exports.take = take;
/**
 * Get the first element of an array
 */
function head(arr) {
    return arr[0];
}
exports.head = head;
/**
 * Get multiple object or array value by paths. Example path: `a[0].b.c`
 */
function at(source, paths) {
    const out = [];
    if (typeof paths === "string") {
        paths = [paths];
    }
    for (let i = 0; i < paths.length; i++) {
        const path = paths[i];
        const parts = path.split(/[.[\]]/g);
        let haystack = source;
        let found = true;
        for (let i = 0; i < parts.length; i++) {
            const part = parts[i];
            if (part === "")
                continue;
            if (!(part in haystack)) {
                found = false;
                break;
            }
            haystack = haystack[part];
        }
        if (found) {
            out.push(haystack);
        }
    }
    return out;
}
exports.at = at;
/**
 * Return item by path or default value if not present
 */
function get(source, path, defaultValue) {
    if (Array.isArray(path)) {
        path = path.join(".");
    }
    const parts = path.split(/[.[\]]/g);
    let haystack = source;
    for (let i = 0; i < parts.length; i++) {
        const part = parts[i];
        if (part === "")
            continue;
        if (!(part in haystack))
            return defaultValue;
        haystack = haystack[part];
    }
    return haystack;
}
exports.get = get;
/**
 * Check if an item is in an object by path
 */
function has(source, path) {
    const empty = "__tlodash_not_found";
    return get(source, path, empty) !== empty;
}
exports.has = has;
/**
 * Check if a collection (Array, Map, Set, Object) is empty
 */
function isEmpty(v) {
    if (Array.isArray(v)) {
        return v.length === 0;
    }
    else if (v instanceof Set || v instanceof Map) {
        return v.size === 0;
    }
    else if (v !== null && typeof v === "object") {
        return Object.keys(v).length === 0;
    }
    return true;
}
exports.isEmpty = isEmpty;
/**
 * Pick values out of an object by predicate and return them in a new object
 */
function pickBy(obj, predicate) {
    const out = {};
    Object.keys(obj).forEach(key => {
        if (predicate(obj[key], key)) {
            out[key] = obj[key];
        }
    });
    return out;
}
exports.pickBy = pickBy;
/**
 * Loop over a collection and call the callback on each item
 */
function forEach(value, callback) {
    if (Array.isArray(value)) {
        value.forEach(callback);
    }
    else if (value instanceof Set || value instanceof Map) {
        value.forEach(callback);
    }
    else if (value !== null && typeof value === "object") {
        Object.keys(value).forEach(key => {
            callback(value[key], key, value);
        });
    }
}
exports.forEach = forEach;
/**
 * Convert string to kebab case
 */
function kebabCase(input) {
    let out = input
        .replace(/[-._\s]+/g, "-")
        .replace(/[A-Z0-9]/, "-$&")
        .replace(/[-]{2,}/, "");
    if (out[0] === "-") {
        out = out.slice(1);
    }
    if (out[out.length - 1] === "-") {
        out = out.slice(0, -1);
    }
    return out.toLowerCase();
}
exports.kebabCase = kebabCase;
/**
 * Get index of an item in an array
 */
function indexOf(arr, value, startIndex = 0) {
    if (startIndex === 0)
        return arr.indexOf(value);
    for (let i = startIndex; i < arr.length; i++) {
        if (arr[i] === value)
            return i;
    }
    return -1;
}
exports.indexOf = indexOf;
/**
 * Flatten an array by a single level
 */
function flatten(arr) {
    return arr.reduce((acc, val) => acc.concat(val), []);
}
exports.flatten = flatten;
/**
 * Remove falsy values from array
 */
function compact(arr) {
    return arr.filter(x => !!x);
}
exports.compact = compact;
/** Used to generate unique IDs. */
const idCounter = {};
/**
 * Generates a unique ID
 */
function uniqueId(prefix = "$smoldash$") {
    const id = (idCounter[prefix] = (idCounter[prefix] || 0) + 1);
    return "" + (prefix === "$smoldash$" ? "" : prefix) + id;
}
exports.uniqueId = uniqueId;
// eslint-disable-next-line @typescript-eslint/ban-types
function once(fn) {
    let result;
    let called = false;
    return function once_wrapped(...args) {
        if (!called) {
            called = true;
            result = fn.apply(this, args);
        }
        return result;
    };
}
exports.once = once;
/**
 * Create a duplicate free version of an array
 */
function uniq(arr) {
    if (!Array.isArray(arr))
        return [];
    return Array.from(new Set(arr).values());
}
exports.uniq = uniq;
/**
 * Create a duplicate free version of an array by a user iteratee
 */
function uniqBy(arr, iteratee) {
    if (!Array.isArray(arr))
        return [];
    const seen = new Map();
    const type = typeof iteratee;
    arr.forEach(item => {
        let key = iteratee;
        if (type === "string" && item !== null && typeof item === "object") {
            key = item[iteratee];
        }
        else if (type === "function") {
            key = iteratee(item);
        }
        if (!seen.has(key)) {
            seen.set(key, item);
        }
    });
    return Array.from(seen.values());
}
exports.uniqBy = uniqBy;
/**
 * Pass the result of the first function to the next one.
 */
function flow(...fns) {
    return function (...args) {
        let result = fns[0].apply(this, args);
        for (let i = 1; i < fns.length; i++) {
            result = fns[i].call(this, result);
        }
        return result;
    };
}
exports.flow = flow;
/**
 * Shallow clone of a value
 */
function clone(value) {
    if (Array.isArray(value)) {
        return value.slice();
    }
    else if (value instanceof RegExp) {
        return new RegExp(value.source, value.flags);
    }
    else if (value instanceof Set) {
        return new Set(value);
    }
    else if (value instanceof Map) {
        return new Map(value);
    }
    else if (value instanceof Date) {
        return new Date(value);
    }
    else if (typeof value === "object" && value !== null) {
        return Object.assign({}, value);
    }
    return value;
}
exports.clone = clone;
/**
 * Deeply clone a value
 */
function cloneDeep(value) {
    if (Array.isArray(value)) {
        return value.slice().map(cloneDeep);
    }
    else if (value instanceof RegExp) {
        return new RegExp(value.source, value.flags);
    }
    else if (value instanceof Set) {
        const out = new Set();
        value.forEach(v => out.add(cloneDeep(v)));
        return out;
    }
    else if (value instanceof Map) {
        const out = new Map();
        value.forEach((v, k) => out.set(k, cloneDeep(v)));
        return out;
    }
    else if (value instanceof Date) {
        return new Date(value);
    }
    else if (typeof value === "object" && value !== null) {
        const out = {};
        for (const k in value) {
            out[k] = cloneDeep(value[k]);
        }
        return out;
    }
    return value;
}
exports.cloneDeep = cloneDeep;
/**
 * Deeply merge multiple objects
 */
function merge(...objs) {
    const a = objs[0];
    for (let i = 1; i < objs.length; i++) {
        const b = objs[i];
        for (const k in b) {
            if (a[k] !== null &&
                b[k] !== null &&
                typeof a[k] === "object" &&
                typeof b[k] === "object") {
                merge(a[k], b[k]);
            }
            else {
                a[k] = b[k];
            }
        }
    }
    return objs[0];
}
exports.merge = merge;
const createPredicate = (predicate = x => !!x) => {
    let fn;
    if (typeof predicate === "string") {
        fn = (item) => !!item[predicate];
    }
    else if (Array.isArray(predicate)) {
        fn = (item) => item[predicate[0]] === predicate[1];
    }
    else if (typeof predicate === "object") {
        fn = (item) => {
            return Object.keys(predicate).every(v => !(v in predicate) || item[v] === predicate[v]);
        };
    }
    else if (typeof predicate === "function") {
        fn = predicate;
    }
    return fn;
};
/**
 * Iterate the collection and return the index of the element where the predicate returns true
 */
function findIndex(collection, predicate = x => !!x, fromIndex = 0) {
    if (!Array.isArray(collection))
        return -1;
    const fn = createPredicate(predicate);
    for (let i = fromIndex; i < collection.length; i++) {
        if (fn(collection[i], i, collection)) {
            return i;
        }
    }
    return -1;
}
exports.findIndex = findIndex;
/**
 * Iterate the collection and return the element where the predicate returns true
 */
function find(collection, predicate = x => !!x, fromIndex = 0) {
    const index = findIndex(collection, predicate, fromIndex);
    return index < 0 ? undefined : collection[index];
}
exports.find = find;
/**
 * Iterate the collection and return the elements where the predicate returns true
 */
function filter(collection, predicate = x => !!x) {
    if (!Array.isArray(collection))
        return [];
    return collection.filter(createPredicate(predicate));
}
exports.filter = filter;
/**
 * Iterate the collection and return true if predicate returns true for at least one element
 */
function some(collection, predicate = x => !!x) {
    return findIndex(collection, predicate) > -1;
}
exports.some = some;
/**
 * Iterate the collection and return true if predicate returns true for all elements
 */
function every(collection, predicate = x => !!x) {
    const fn = createPredicate(predicate);
    // if we find one element which does not satisfy predicate, then return false
    return findIndex(collection, item => !fn(item)) < 0;
}
exports.every = every;
function map(collection, iteratee = x => x) {
    if (!Array.isArray(collection))
        return [];
    let fn;
    if (typeof iteratee === "string") {
        fn = (item) => item[iteratee];
    }
    else if (typeof iteratee === "function") {
        fn = iteratee;
    }
    return collection.map(fn);
}
exports.map = map;
/**
 * Shallowly compare to values
 */
function isEqual(a, b) {
    if (a === null || b === null) {
        return a === b;
    }
    else if (Array.isArray(a) && Array.isArray(b)) {
        if (a.length !== b.length) {
            return false;
        }
        for (let i = 0; i < a.length; i++) {
            if (a[i] !== b[i]) {
                return false;
            }
        }
        return true;
    }
    else if (a instanceof RegExp && b instanceof RegExp) {
        return "" + a === "" + b;
    }
    else if (a instanceof Set && b instanceof Set) {
        if (a.size !== b.size)
            return false;
        for (const v of a.values()) {
            if (!b.has(v))
                return false;
        }
        return true;
    }
    else if (a instanceof Map && b instanceof Map) {
        if (a.size !== b.size)
            return false;
        for (const [k, v] of a.entries()) {
            if (!b.has(k))
                return false;
            if (b.get(k) !== v)
                return false;
        }
        return true;
    }
    else if (a instanceof Date && b instanceof Date) {
        return +a === +b;
    }
    else if (typeof a === "object" && typeof b === "object") {
        for (const i in a)
            if (!(i in b))
                return false;
        for (const i in b)
            if (a[i] !== b[i])
                return false;
        return true;
    }
    return a === b;
}
exports.isEqual = isEqual;
/**
 * Generate a range of numbers
 */
function range(start, end, step = 1) {
    if (typeof end === "undefined") {
        end = start;
        start = 0;
    }
    const result = [];
    for (let i = start; i < end; i += step) {
        result.push(i);
    }
    return result;
}
exports.range = range;
/**
 * Creates object to items in collection using keys from iteratee
 */
function keyBy(collection, iteratee) {
    if (!Array.isArray(collection))
        return {};
    let fn;
    if (typeof iteratee === "string") {
        fn = (item) => item[iteratee];
    }
    else if (typeof iteratee === "function") {
        fn = iteratee;
    }
    return collection.reduce((accumulator, item, index) => {
        const key = fn(item, index, collection);
        accumulator[key] = item;
        return accumulator;
    }, {});
}
exports.keyBy = keyBy;
/**
 * Creates object to items in collection using keys from iteratee
 */
function groupBy(collection, iteratee) {
    if (!Array.isArray(collection))
        return {};
    let fn;
    if (typeof iteratee === "string") {
        fn = (item) => item[iteratee];
    }
    else if (typeof iteratee === "function") {
        fn = iteratee;
    }
    return collection.reduce((accumulator, item, index) => {
        const key = fn(item, index, collection);
        if (!accumulator[key])
            accumulator[key] = [];
        accumulator[key].push(item);
        return accumulator;
    }, {});
}
exports.groupBy = groupBy;
